from django.core import management
from django.conf import settings
from django.db.utils import IntegrityError
from django.core.management.base import BaseCommand, CommandError
from django.core.exceptions import ValidationError

class Command(BaseCommand):

    help = 'Initialize the DjangoLDP backend'

    def add_arguments(self, parser):

        """Define the same arguments as the ones in CLI."""

        parser.add_argument('--with-admin', nargs='?', type=str, help='Create an administrator user.')
        parser.add_argument('--email', nargs='?', type=str, help='Provide an email for administrator.')
        parser.add_argument('--with-dummy-admin', action='store_true', help='Create a default "admin" user with "admin" password.')

    def handle(self, *args, **options):

        """Wrapper command around default django initialization commands."""

        try:
            # migrate data
            management.call_command('migrate', interactive=False)

        except CommandError as e:
            self.stdout.write(self.style.ERROR(f'Data migration failed: {e}'))

        if options['with_dummy_admin']:
            try:
                # create a default super user
                from django.contrib.auth import get_user_model
                User = get_user_model()
                User.objects.create_superuser('admin', 'admin@example.org', 'admin')

            except (ValidationError, IntegrityError):
                self.stdout.write('User "admin" already exists. Skipping...')
                pass

        elif options['with_admin']:
            try:
                # call default createsuperuser command
                management.call_command('createsuperuser', '--noinput', '--username', options['with_admin'], '--email', options['email'])

            except CommandError as e:
                self.stdout.write(self.style.ERROR(f'Superuser {e}'))
                pass

        #try:
        #    # creatersakey
        #    management.call_command('creatersakey', interactive=False)

        #except CommandError as e:
        #    self.stdout.write(self.style.ERROR(f'RSA key creation failed: {e}'))



"""This module override some of default django configuration from the global settings."""

from django.conf.global_settings import *

####################
# LDP              #
####################

LDP_RDF_CONTEXT = 'https://cdn.happy-dev.fr/owl/hdcontext.jsonld'

MAX_ACTIVITY_RESCHEDULES = 3
DEFAULT_BACKOFF_FACTOR = 1
DEFAULT_ACTIVITY_DELAY = 3
DEFAULT_REQUEST_TIMEOUT = 10

####################
# CORE             #
####################

# https://en.wikipedia.org/wiki/List_of_tz_zones_by_name (although not all
# systems may support all possibilities). When USE_TZ is True, this is
# interpreted as the default user time zone.
TIME_ZONE = 'UTC'

# If you set this to True, Django will use timezone-aware datetimes.
USE_TZ = True

# Language code for this installation. All choices can be found here:
# http://www.i18nguy.com/unicode/language-identifiers.html
LANGUAGE_CODE = 'en-us'

# Database connection info. If left empty, will default to the dummy backend.
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': 'db.sqlite3',
    }
}

# List of strings representing installed apps.
INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'djangoldp',
    'guardian'
]

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

# Default server url
SITE_URL = 'http://localhost:8000'
BASE_URL = SITE_URL

# Default URL conf
ROOT_URLCONF = 'djangoldp.urls'

# Absolute filesystem path to the directory that will hold user-uploaded files.
# Example: "/var/www/example.com/media/"
MEDIA_ROOT = ''

# URL that handles the media served from MEDIA_ROOT.
# Examples: "http://example.com/media/", "http://media.example.com/"
MEDIA_URL = '/media/'

# Absolute path to the directory static files should be collected to.
# Example: "/var/www/example.com/static/"
STATIC_ROOT = None

# URL that handles the static files served from STATIC_ROOT.
# Example: "http://example.com/static/", "http://static.example.com/"
STATIC_URL = '/static/'

# ETAGS config
USE_ETAGS = True

# Default X-Frame-Options header value
X_FRAME_OPTIONS = 'DENY'

# The Python dotted path to the WSGI application that Django's internal server
# (runserver) will use. If `None`, the return value of
# 'django.core.wsgi.get_wsgi_application' is used, thus preserving the same
# behavior as previous versions of Django. Otherwise this should point to an
# actual WSGI application object.
WSGI_APPLICATION = 'server.wsgi.application'

##############
# MIDDLEWARE #
##############

# List of middleware to use. Order is important; in the request phase, these
# middleware will be applied in the order given, and in the response
# phase the middleware will be applied in reverse order.
MIDDLEWARE = [
    'django_cookies_samesite.middleware.CookiesSameSite',
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'django.middleware.gzip.GZipMiddleware',
    'django_brotli.middleware.BrotliMiddleware'
]

############
# SESSIONS #
############

# Same site policy
DCS_SESSION_COOKIE_SAMESITE = 'none'

##################
# AUTHENTICATION #
##################

AUTHENTICATION_BACKENDS = ['django.contrib.auth.backends.ModelBackend', 'guardian.backends.ObjectPermissionBackend']

OIDC_ACCESS_CONTROL_ALLOW_HEADERS = 'Content-Type, if-match, accept, authorization, DPoP'

# The minimum number of seconds a password reset link is valid for
PASSWORD_RESET_TIMEOUT = 60 * 60 * 24 * 3




from importlib import import_module

from django.conf import settings
from django.conf.urls import re_path, include

from djangoldp.models import LDPSource, Model
from djangoldp.permissions import LDPPermissions
from djangoldp.views import LDPSourceViewSet, WebFingerView, InboxView
from djangoldp.views import LDPViewSet


def __clean_path(path):
    '''ensures path is Django-friendly'''
    if path.startswith("/"):
        path = path[1:]
    if not path.endswith("/"):
        path = "{}/".format(path)
    return path


def get_all_non_abstract_subclasses(cls):
    '''
    returns a set of all subclasses for a given Python class (recursively calls cls.__subclasses__()). Ignores Abstract
    classes
    '''
    def valid_subclass(sc):
        '''returns True if the parameterised subclass is valid and should be returned'''
        return not Model.get_meta(sc, 'abstract', False)

    return set(c for c in cls.__subclasses__() if valid_subclass(c)).union(
        [s for c in cls.__subclasses__() for s in get_all_non_abstract_subclasses(c) if valid_subclass(s)])


def get_all_non_abstract_subclasses_dict(cls):
    '''returns a dict of class name -> class for all subclasses of given cls parameter (recursively)'''
    return {cls.__name__: cls for cls in get_all_non_abstract_subclasses(cls)}


urlpatterns = [
    re_path(r'^sources/(?P<federation>\w+)/', LDPSourceViewSet.urls(model=LDPSource, fields=['federation', 'urlid'],
                                                                    permission_classes=[LDPPermissions], )),
    re_path(r'^\.well-known/webfinger/?$', WebFingerView.as_view()),
    re_path(r'^inbox/$', InboxView.as_view()),
]

for package in settings.DJANGOLDP_PACKAGES:
    try:
        import_module('{}.models'.format(package))
        urlpatterns.append(re_path(r'^', include('{}.djangoldp_urls'.format(package))))
    except ModuleNotFoundError:
        pass

# fetch a list of all models which subclass DjangoLDP Model
model_classes = get_all_non_abstract_subclasses_dict(Model)

# append urls for all DjangoLDP Model subclasses
for class_name in model_classes:
    model_class = model_classes[class_name]
    # the path is the url for this model
    path = __clean_path(model_class.get_container_path())
    # urls_fct will be a method which generates urls for a ViewSet (defined in LDPViewSetGenerator)
    urls_fct = model_class.get_view_set().urls
    urlpatterns.append(re_path(r'^' + path,
        urls_fct(model=model_class,
                 lookup_field=Model.get_meta(model_class, 'lookup_field', 'pk'),
                 permission_classes=Model.get_meta(model_class, 'permission_classes', [LDPPermissions]),
                 fields=Model.get_meta(model_class, 'serializer_fields', []),
                 nested_fields=model_class.nested.fields())))

# NOTE: this route will be ignored if a custom (subclass of Model) user model is used, or it is registered by a package
# Django matches the first url it finds for a given path
urlpatterns.append(re_path(r'^users/', LDPViewSet.urls(model=settings.AUTH_USER_MODEL, permission_classes=[])))



from django.contrib import admin

from djangoldp_crypto.models import RSAKey


@admin.register(RSAKey)
class RSAKeyAdmin(admin.ModelAdmin):

    readonly_fields = ['kid', 'pub_key']

    def save_model(self, request, obj, form, change):
        obj.priv_key.replace('\r', '')
        super().save_model(request, obj, form, change)



import argparse
from django.core.management.base import BaseCommand
from django.conf import settings
from djangoldp.models import LDPSource

def isstring(target):
  if isinstance(target, str):
    return target
  return False

def list_models():
  # Improve me using apps.get_models()
  return {
    "circles": "/circles/",
    "circlesjoinable": "/circles/joinable/",
    "communities": "/communities/",
    "opencommunities": "/open-communities/",
    "communitiesaddresses": "/communities-addresses/",
    "dashboards": "/dashboards/",
    "events": "/events/",
    "eventsfuture": "/events/future/",
    "eventspast": "/events/past/",
    "typeevents": "/typeevents/",
    "resources": "/resources/",
    "keywords": "/keywords/",
    "types": "/types/",
    "joboffers": "/job-offers/current/",
    "projects": "/projects/",
    "projectsjoinable": "/projects/joinable/",
    "users": "/users/"
  }

class Command(BaseCommand):
  help = 'Add another server to this one sources'

  def add_arguments(self, parser):
    parser.add_argument(
      "--target",
      action="store",
      default=False,
      type=isstring,
      help="Targeted server, format protocol://domain",
    )
    parser.add_argument(
      "--delete",
      default=False,
      nargs='?',
      const=True,
      help="Remove targeted source",
    )

  def handle(self, *args, **options):
    target = options["target"]
    models = list_models()
    if not target:
      target = settings.SITE_URL
    error_counter = 0
    if(options["delete"]):
      for attr, value in models.items():
        try:
          LDPSource.objects.filter(urlid=target+value, federation=attr).delete()
        except:
          error_counter += 1
      if error_counter > 0:
        self.stdout.write(self.style.ERROR("Can't remove: "+target))
        exit(2)
      else:
        self.stdout.write(self.style.SUCCESS("Successfully removed sources for "+target))
        exit(0)
    else:
      for attr, value in models.items():
        try:
          LDPSource.objects.create(urlid=target+value, federation=attr)
        except:
          error_counter += 1
      if error_counter > 0:
        self.stdout.write(self.style.WARNING("Source aleady exists for "+target+"\nIgnored "+str(error_counter)+"/"+str(len(models))))
        if error_counter == len(models):
          exit(1)
        exit(0)
      else:
        self.stdout.write(self.style.SUCCESS("Successfully created sources for "+target))
        exit(0)



from django.conf import settings
from guardian.utils import get_anonymous_user


# convenience function returns True if user is anonymous
def is_anonymous_user(user):
    return user.is_anonymous or (getattr(settings, 'ANONYMOUS_USER_NAME', True) is not None and
                                 user == get_anonymous_user())


# convenience function returns True if user is authenticated
def is_authenticated_user(user):
    return user.is_authenticated and (getattr(settings, 'ANONYMOUS_USER_NAME', True) is None or
                                      user != get_anonymous_user())



import uuid
from django.conf import settings
from django.contrib.auth import get_user_model
from django.test import TestCase, override_settings
from rest_framework.test import APIRequestFactory, APIClient

from djangoldp.serializers import LDPSerializer, LDListMixin
from djangoldp.tests.models import Post, UserProfile, Resource, Circle, CircleMember, Invoice, Batch, Task, ModelTask
from djangoldp.tests.models import Skill, JobOffer, Conversation, Message, Project


class LDPModelSerializerTestCase(TestCase):
    def setUp(self):
        self.factory = APIRequestFactory()
        self.client = APIClient()
        self.user = get_user_model().objects.create_user(username='john', email='jlennon@beatles.com',
                                                         password='glass onion')
        self.client.force_authenticate(user=self.user)
        LDListMixin.to_representation_cache.reset()
        LDPSerializer.to_representation_cache.reset()

    def _get_serializer_class(self, model, depth, fields):
        meta_args = {'model': model, 'depth': depth, 'fields': fields}

        meta_class = type('Meta', (), meta_args)
        return type(LDPSerializer)('TestSerializer', (LDPSerializer,), {'Meta': meta_class})

    def test_update_container_new_resource_replace(self):
        # 2 pre-existing skills, one will be replaced and the other updated
        redundant_skill = Skill.objects.create(title="to drop", obligatoire="obligatoire", slug="slug1")
        pre_existing_skill = Skill.objects.create(title="to keep", obligatoire="obligatoire", slug="slug2")
        job = JobOffer.objects.create(title="job test")
        job.skills.add(redundant_skill)
        job.skills.add(pre_existing_skill)

        post = {"@id": "{}/job-offers/{}/".format(settings.BASE_URL, job.slug),
               "title": "job test updated",
               "skills": {
                   "ldp:contains": [
                       {"title": "new skill", "obligatoire": "okay"},
                       {"@id": "{}/skills/{}/".format(settings.BASE_URL, pre_existing_skill.slug), "title": "z"},
                   ]}
               }

        serializer_class = self._get_serializer_class(JobOffer, 2, ("@id", "title", "skills"))
        serializer = serializer_class(data=post, instance=job)
        serializer.is_valid(raise_exception=True)
        result = serializer.save()

        self.assertEquals(result.title, "job test updated")
        self.assertIs(result.skills.count(), 2)
        skills = result.skills.all().order_by("title")
        self.assertEquals(skills[0].title, "new skill")
        self.assertEquals(skills[0].obligatoire, "okay")
        self.assertEquals(skills[1].title, "z") # updated
        self.assertEquals(skills[1].obligatoire, pre_existing_skill.obligatoire)

    # TODO: https://git.startinblox.com/djangoldp-packages/djangoldp/issues/326
    '''
    def test_update_container_edit_and_new_resource_append(self):
        pre_existing_skill_a = Skill.objects.create(title="to keep", obligatoire="obligatoire", slug="slug1")
        pre_existing_skill_b = Skill.objects.create(title="to keep", obligatoire="obligatoire", slug="slug2")
        job = JobOffer.objects.create(title="job test")
        job.skills.add(pre_existing_skill_a)
        job.skills.add(pre_existing_skill_b)

        post = {"@id": "{}/job-offers/{}/".format(settings.BASE_URL, job.slug),
                "skills": {
                    "ldp:contains": [
                        {"title": "new skill", "obligatoire": "okay"},
                        {"@id": "{}/skills/{}/".format(settings.BASE_URL, pre_existing_skill_b.slug), "title": "z"},
                    ]}
                }

        serializer_class = self._get_serializer_class(JobOffer, 2, ("@id", "title", "skills"))
        serializer = serializer_class(data=post, instance=job)
        serializer.is_valid(raise_exception=True)
        result = serializer.save(partial=True)

        self.assertEquals(result.title, job.title)
        self.assertIs(result.skills.count(), 3)
        skills = result.skills.all().order_by('title')
        self.assertEquals(skills[0].title, "new skill") # new skill
        self.assertEquals(skills[1].title, pre_existing_skill_a.title) # old skill unchanged
        self.assertEquals(skills[2].title, "z") # updated
        self.assertEquals(skills[2].obligatoire, pre_existing_skill_b.obligatoire) # another field not updated
    '''

    def test_update_container_edit_and_new_external_resources(self):
        job = JobOffer.objects.create(title="job test")
        pre_existing_external = Skill.objects.create(title="to keep", obligatoire="obligatoire",
                                                     urlid="https://external.com/skills/2/")
        job.skills.add(pre_existing_external)

        post = {"@id": "{}/job-offers/{}/".format(settings.BASE_URL, job.slug),
                "skills": {
                    "ldp:contains": [
                        {"@id": "https://external.com/skills/1/", "title": "external skill", "obligatoire": "okay"},
                        {"@id": "https://external.com/skills/2/", "title": "to keep", "obligatoire": "okay"},
                    ]}
                }

        serializer_class = self._get_serializer_class(JobOffer, 2, ("@id", "title", "skills"))
        serializer = serializer_class(data=post, instance=job)
        serializer.is_valid(raise_exception=True)
        result = serializer.save()

        skills = result.skills.all().order_by('urlid')
        self.assertEquals(result.title, job.title)
        self.assertEqual(result.pk, job.pk)
        self.assertEqual(result.urlid, job.urlid)
        self.assertIs(result.skills.count(), 2)
        self.assertEquals(skills[0].title, "external skill")  # new skill
        self.assertEquals(skills[0].urlid, "https://external.com/skills/1/")  # new skill
        self.assertEquals(skills[0].obligatoire, "okay")
        self.assertEquals(skills[1].title, pre_existing_external.title)  # old skill unchanged
        self.assertEquals(skills[1].urlid, pre_existing_external.urlid)
        self.assertEquals(skills[1].obligatoire, "okay")
        self.assertEquals(skills[1].pk, pre_existing_external.pk)

    def test_update_container_attach_existing_resource(self):
        job = JobOffer.objects.create(title="job test")
        another_job = JobOffer.objects.create(title="job2")
        pre_existing_skill = Skill.objects.create(title="to keep", obligatoire="obligatoire")
        another_job.skills.add(pre_existing_skill)

        self.assertIs(job.skills.count(), 0)

        post = {"@id": "{}/job-offers/{}/".format(settings.BASE_URL, job.slug),
                "skills": {
                    "ldp:contains": [
                        {"@id": "{}/skills/{}/".format(settings.BASE_URL, pre_existing_skill.slug)},
                    ]}
                }

        serializer_class = self._get_serializer_class(JobOffer, 2, ("@id", "title", "skills"))
        serializer = serializer_class(data=post, instance=job)
        serializer.is_valid(raise_exception=True)
        result = serializer.save()

        skills = result.skills.all().order_by('urlid')
        self.assertEquals(result.title, job.title)
        self.assertEqual(result.pk, job.pk)
        self.assertEqual(result.urlid, job.urlid)
        self.assertIs(result.skills.count(), 1)
        self.assertEquals(skills[0].urlid, pre_existing_skill.urlid)
        self.assertIs(another_job.skills.count(), 1)
        self.assertIs(Skill.objects.count(), 1)

    def test_update_container_attach_existing_resource_external(self):
        job = JobOffer.objects.create(title="job test")
        another_job = JobOffer.objects.create(title="job2")
        pre_existing_external = Skill.objects.create(title="to keep", obligatoire="obligatoire",
                                                     urlid="https://external.com/skills/2/")
        another_job.skills.add(pre_existing_external)

        self.assertIs(job.skills.count(), 0)

        post = {"@id": "{}/job-offers/{}/".format(settings.BASE_URL, job.slug),
                "skills": {
                    "ldp:contains": [
                        {"@id": pre_existing_external.urlid},
                    ]}
                }

        serializer_class = self._get_serializer_class(JobOffer, 2, ("@id", "title", "skills"))
        serializer = serializer_class(data=post, instance=job)
        serializer.is_valid(raise_exception=True)
        result = serializer.save()

        skills = result.skills.all().order_by('urlid')
        self.assertEquals(result.title, job.title)
        self.assertEqual(result.pk, job.pk)
        self.assertEqual(result.urlid, job.urlid)
        self.assertIs(result.skills.count(), 1)
        self.assertEquals(skills[0].urlid, pre_existing_external.urlid)
        self.assertIs(another_job.skills.count(), 1)
        self.assertIs(Skill.objects.count(), 1)

    # TODO: https://git.startinblox.com/djangoldp-packages/djangoldp/issues/344
    def test_update_container_mismatched_type_urlid(self):
        job = JobOffer.objects.create(title="job test")
        another_job = JobOffer.objects.create(title="job2")

        # contains internal urlid which refers to a different type of object entirely, and one which refers to container
        post = {"@id": "{}/job-offers/{}/".format(settings.BASE_URL, job.slug),
                "skills": {
                    "ldp:contains": [
                        {"@id": "{}/job-offers/{}/".format(settings.BASE_URL, another_job.slug)},
                    ]}
                }

        serializer_class = self._get_serializer_class(JobOffer, 2, ("@id", "title", "skills"))
        serializer = serializer_class(data=post, instance=job)
        serializer.is_valid(raise_exception=True)
        result = serializer.save()

    # TODO: https://git.startinblox.com/djangoldp-packages/djangoldp/issues/345
    '''
    def test_update_container_mismatched_type_urlid_2(self):
        job = JobOffer.objects.create(title="job test")

        # contains internal urlid which refers to a container
        post = {"@id": "{}/job-offers/{}/".format(settings.BASE_URL, job.slug),
                "skills": {
                    "ldp:contains": [
                        {"@id": "{}/skills/".format(settings.BASE_URL)},
                    ]}
                }

        serializer_class = self._get_serializer_class(JobOffer, 2, ("@id", "title", "skills"))
        serializer = serializer_class(data=post, instance=job)
        serializer.is_valid(raise_exception=True)
        result = serializer.save()

        # TODO: assert correct error is thrown
    '''

    # TODO: https://git.startinblox.com/djangoldp-packages/djangoldp/issues/344
    def test_update_container_mismatched_type_urlid_external(self):
        job = JobOffer.objects.create(title="job test")

        # contains external mismatched urlids which refers to a container
        post = {"@id": "{}/job-offers/{}/".format(settings.BASE_URL, job.slug),
                "skills": {
                    "ldp:contains": [
                        {"@id": "https://external.com/skills/"},
                    ]}
                }

        serializer_class = self._get_serializer_class(JobOffer, 2, ("@id", "title", "skills"))
        serializer = serializer_class(data=post, instance=job)
        serializer.is_valid(raise_exception=True)
        result = serializer.save()

    # TODO: https://git.startinblox.com/djangoldp-packages/djangoldp/issues/346
    '''def test_update_container_attach_nonexistent_local_resource(self):
        job = JobOffer.objects.create(title="job test")

        self.assertEqual(JobOffer.objects.count(), 1)
        self.assertEqual(job.skills.count(), 0)
        self.assertEqual(Skill.objects.count(), 0)

        post = {"@id": "{}/job-offers/{}/".format(settings.BASE_URL, job.slug),
                "skills": {
                    "ldp:contains": [
                        {"@id": "{}/skills/404/".format(settings.BASE_URL)},
                    ]}
                }

        serializer_class = self._get_serializer_class(JobOffer, 2, ("@id", "title", "skills"))
        serializer = serializer_class(data=post, instance=job)
        serializer.is_valid(raise_exception=True)
        result = serializer.save()

        self.assertEqual(JobOffer.objects.count(), 1)
        self.assertEqual(job.skills.count(), 0)
        self.assertEqual(Skill.objects.count(), 0)'''

    # CircleMember is different to Skill because it represents a many-to-many relationship via a through model
    # TODO: https://git.startinblox.com/djangoldp-packages/djangoldp/issues/333
    '''def test_update_m2m_relationship_with_through_model_add_and_edit(self):
        circle = Circle.objects.create(name='test')
        pre_existing = CircleMember.objects.create(user=self.user, circle=circle, is_admin=False)
        another_user = get_user_model().objects.create_user(username='u2', email='u2@b.com', password='pw')

        post = {
            "@id": "{}/circles/{}/".format(settings.BASE_URL, circle.pk),
            "name": "Updated Name",
            "members": {
                "ldp:contains": [
                    {"@id": "{}/circle-members/{}/".format(settings.BASE_URL, pre_existing.pk), "is_admin": True},
                    {"user": {"@id": another_user.urlid }, "is_admin": False},
                ]
            }
        }

        serializer_class = self._get_serializer_class(Circle, 2, ("@id", "name", "description", "members", "team"))
        serializer = serializer_class(data=post, instance=circle)
        serializer.is_valid(raise_exception=True)
        result = serializer.save()

        self.assertEquals(result.name, circle.name)
        self.assertEqual(result.pk, circle.pk)
        self.assertEqual(result.urlid, circle.urlid)
        self.assertIs(result.members.count(), 2)
        self.assertIs(result.team.count(), 2)

        members = result.members.all().order_by('pk')
        self.assertEqual(members[0].user, self.user)
        self.assertEqual(members[0].urlid, pre_existing.urlid)
        self.assertEqual(members[0].pk, pre_existing.pk)
        self.assertEqual(members[0].is_admin, True)
        self.assertEqual(members[1].user, another_user)
        self.assertEqual(members[1].is_admin, False)

    # TODO: variation on the above using external resources
    def test_update_m2m_relationship_with_through_model_add_and_edit_external_resources(self):
        pass

    # NOTE: this test if failing due to missing the 'invoice_id' field (see #333)
    #  variation of this test exists in tests_update.py with different behaviour
    def test_update_container_twice_nested(self):
        invoice = Invoice.objects.create(title='test')
        pre_existing_batch = Batch.objects.create(title='batch1', invoice=invoice)
        pre_existing_task = ModelTask.objects.create(title='task1', batch=pre_existing_batch)

        post = {
          "@id": "{}/invoices/{}/".format(settings.BASE_URL, invoice.pk),
          "title": "new",
          "batches": [
            {
              "@id": "{}/batchs/{}/".format(settings.BASE_URL, pre_existing_batch.pk),
              "title": "new",
              "tasks": [
                {
                  "@id": "{}/modeltasks/{}/".format(settings.BASE_URL, pre_existing_task.pk),
                  "title": "new"
                },
                {
                  "title": "tache 2"
                }
              ]
            },
            {
              "title": "z",
            }
          ]
        }

        serializer_class = self._get_serializer_class(Invoice, 2, ("@id", "title", "batches"))
        serializer = serializer_class(data=post, instance=invoice)
        serializer.is_valid(raise_exception=True)
        result = serializer.save()

        self.assertEquals(result.title, "new")
        self.assertEquals(result.urlid, invoice.urlid)
        self.assertEquals(result.pk, invoice.pk)

        self.assertIs(result.batches.count(), 2)
        batches = result.batches.all().order_by('title')
        self.assertEquals(batches[0].title, "new")
        self.assertEquals(batches[0].urlid, pre_existing_batch.urlid)
        self.assertEquals(batches[1].title, "z")

        self.assertIs(batches[0].tasks.count(), 2)
        tasks = batches[0].tasks.all().order_by('title')
        self.assertEquals(tasks[0].title, "new")
        self.assertEquals(tasks[0].urlid, pre_existing_task.urlid)
        self.assertEquals(tasks[1].title, "tache 2")

    # variation on the above test with external resources
    def test_update_container_twice_nested_external_resources(self):
        invoice = Invoice.objects.create(urlid='https://external.com/invoices/1/', title='test')
        pre_existing_batch = Batch.objects.create(urlid='https://external.com/batchs/1/', title='batch1', invoice=invoice)
        pre_existing_task = ModelTask.objects.create(urlid='https://external.com/tasks/1/', title='task1', batch=pre_existing_batch)

        post = {
            "@id": invoice.urlid,
            "title": "new",
            "batches": [
                {
                    "@id": pre_existing_batch.urlid,
                    "title": "new",
                    "tasks": [
                        {
                            "@id": pre_existing_task.urlid,
                            "title": "new"
                        },
                        {
                            "@id": "https://anotherexternal.com/tasks/1/",
                            "title": "tache 2"
                        }
                    ]
                },
                {
                    "@id": "https://yetanotherexternal.com/batchs/1/",
                    "title": "z"
                }
            ]
        }

        serializer_class = self._get_serializer_class(Invoice, 2, ("@id", "title", "batches"))
        serializer = serializer_class(data=post, instance=invoice)
        serializer.is_valid(raise_exception=True)
        result = serializer.save()

        self.assertEquals(result.title, "new")
        self.assertEquals(result.urlid, invoice.urlid)
        self.assertEquals(result.pk, invoice.pk)

        self.assertIs(result.batches.count(), 2)
        batches = result.batches.all().order_by('title')
        self.assertEquals(batches[0].title, "new")
        self.assertEquals(batches[0].urlid, pre_existing_batch.urlid)
        self.assertEquals(batches[1].title, "z")

        self.assertIs(batches[0].tasks.count(), 2)
        tasks = batches[0].tasks.all().order_by('title')
        self.assertEquals(tasks[0].title, "new")
        self.assertEquals(tasks[0].urlid, pre_existing_task.urlid)
        self.assertEquals(tasks[1].title, "tache 2")'''

    # variation on the test where a field is omitted on each level (no changes are made)
    def test_update_container_twice_nested_no_changes_missing_fields(self):
        invoice = Invoice.objects.create(title='test')
        pre_existing_batch = Batch.objects.create(title='batch1', invoice=invoice)
        pre_existing_task = ModelTask.objects.create(title='task1', batch=pre_existing_batch)

        post = {
            "@id": "{}/invoices/{}/".format(settings.BASE_URL, invoice.pk),
            "batches": [
                {
                    "@id": "{}/batchs/{}/".format(settings.BASE_URL, pre_existing_batch.pk),
                    "tasks": [
                        {
                            "@id": "{}/tasks/{}/".format(settings.BASE_URL, pre_existing_task.pk),
                        }
                    ]
                }
            ]
        }

        serializer_class = self._get_serializer_class(Invoice, 2, ("@id", "title", "batches"))
        serializer = serializer_class(data=post, instance=invoice)
        serializer.is_valid(raise_exception=True)
        result = serializer.save(partial=True)

        self.assertEquals(result.title, invoice.title)
        self.assertEquals(result.urlid, invoice.urlid)
        self.assertEquals(result.pk, invoice.pk)

        self.assertIs(result.batches.count(), 1)
        batches = result.batches.all()
        self.assertEquals(batches[0].title, pre_existing_batch.title)
        self.assertEquals(batches[0].urlid, pre_existing_batch.urlid)

        self.assertIs(batches[0].tasks.count(), 1)
        tasks = batches[0].tasks.all()
        self.assertEquals(tasks[0].title, pre_existing_task.title)

    def test_update_graph_edit_and_new_resource(self):
        redundant_skill = Skill.objects.create(title="to drop", obligatoire="obligatoire", slug="slug1")
        skill1 = Skill.objects.create(title="skill1", obligatoire="obligatoire", slug="slug2")
        skill2 = Skill.objects.create(title="skill2", obligatoire="obligatoire", slug="slug3")
        job1 = JobOffer.objects.create(title="job test", slug="slug4")
        job1.skills.add(redundant_skill)

        job = {"@graph":
            [
                {
                    "@id": "{}/job-offers/{}/".format(settings.BASE_URL, job1.slug),
                    "title": "job test updated",
                    "skills": {
                        "ldp:contains": [
                            {"@id": "{}/skills/{}/".format(settings.BASE_URL, skill1.slug)},
                            {"@id": "{}/skills/{}/".format(settings.BASE_URL, skill2.slug)},
                            {"@id": "_.123"},
                        ]}
                },
                {
                    "@id": "_.123",
                    "title": "new skill",
                    "obligatoire": "okay"
                },
                {
                    "@id": "{}/skills/{}/".format(settings.BASE_URL, skill1.slug),
                },
                {
                    "@id": "{}/skills/{}/".format(settings.BASE_URL, skill2.slug),
                    "title": "skill2 UP"
                }
            ]
        }

        serializer_class = self._get_serializer_class(JobOffer, 2, ("@id", "title", "skills"))
        serializer = serializer_class(data=job, instance=job1)
        serializer.is_valid(raise_exception=True)
        result = serializer.save()

        skills = result.skills.all().order_by('title')

        self.assertEquals(result.title, "job test updated")
        self.assertIs(result.skills.count(), 3)
        self.assertEquals(skills[0].title, "new skill")  # new skill
        self.assertEquals(skills[1].title, "skill1")  # no change
        self.assertEquals(skills[2].title, "skill2 UP")  # title updated

    def test_update_graph_2(self):
        skill = Skill.objects.create(title="to drop", obligatoire="obligatoire", slug="slug")
        skill1 = Skill.objects.create(title="skill1", obligatoire="obligatoire", slug="slug1")
        skill2 = Skill.objects.create(title="skill2", obligatoire="obligatoire", slug="slug2")
        job1 = JobOffer.objects.create(title="job test", slug="slug1")
        job1.skills.add(skill)

        job = {"@graph":
            [
                {
                    "@id": "{}/job-offers/{}/".format(settings.BASE_URL, job1.slug),
                    "title": "job test updated",
                    "skills": {
                        "@id": "{}/job-offers/{}/skills/".format(settings.BASE_URL, job1.slug)
                    }
                },
                {
                    "@id": "_.123",
                    "title": "new skill",
                    "obligatoire": "okay"
                },
                {
                    "@id": "{}/skills/{}/".format(settings.BASE_URL, skill1.slug),
                },
                {
                    "@id": "{}/skills/{}/".format(settings.BASE_URL, skill2.slug),
                    "title": "skill2 UP"
                },
                {
                    '@id': "{}/job-offers/{}/skills/".format(settings.BASE_URL, job1.slug),
                    "ldp:contains": [
                        {"@id": "{}/skills/{}/".format(settings.BASE_URL, skill1.slug)},
                        {"@id": "{}/skills/{}/".format(settings.BASE_URL, skill2.slug)},
                        {"@id": "_.123"},
                    ]
                }
            ]
        }

        meta_args = {'model': JobOffer, 'depth': 2, 'fields': ("@id", "title", "skills", "slug")}

        meta_class = type('Meta', (), meta_args)
        serializer_class = type(LDPSerializer)('JobOfferSerializer', (LDPSerializer,), {'Meta': meta_class})
        serializer = serializer_class(data=job, instance=job1)
        serializer.is_valid(raise_exception=True)
        result = serializer.save()

        skills = result.skills.all().order_by('title')

        self.assertEquals(result.title, "job test updated")
        self.assertIs(result.skills.count(), 3)
        self.assertEquals(skills[0].title, "new skill")  # new skill
        self.assertEquals(skills[1].title, "skill1")  # no change
        self.assertEquals(skills[2].title, "skill2 UP")  # title updated
        self.assertEquals(skill, skill._meta.model.objects.get(pk=skill.pk))  # title updated

    def test_update_list_with_reverse_relation(self):
        user1 = get_user_model().objects.create()
        conversation = Conversation.objects.create(description="Conversation 1", author_user=user1)
        message1 = Message.objects.create(text="Message 1", conversation=conversation, author_user=user1)
        message2 = Message.objects.create(text="Message 2", conversation=conversation, author_user=user1)

        json = {"@graph": [
            {
                "@id": "{}/messages/{}/".format(settings.BASE_URL, message1.pk),
                "text": "Message 1 UP"
            },
            {
                "@id": "{}/messages/{}/".format(settings.BASE_URL, message2.pk),
                "text": "Message 2 UP"
            },
            {
                '@id': "{}/conversations/{}/".format(settings.BASE_URL, conversation.pk),
                'description': "Conversation 1 UP",
                "message_set": [
                    {"@id": "{}/messages/{}/".format(settings.BASE_URL, message1.pk)},
                    {"@id": "{}/messages/{}/".format(settings.BASE_URL, message2.pk)},
                ]
            }
        ]
        }

        meta_args = {'model': Conversation, 'depth': 2, 'fields': ("@id", "description", "message_set")}

        meta_class = type('Meta', (), meta_args)
        serializer_class = type(LDPSerializer)('ConversationSerializer', (LDPSerializer,), {'Meta': meta_class})
        serializer = serializer_class(data=json, instance=conversation)
        serializer.is_valid(raise_exception=True)
        result = serializer.save()

        messages = result.message_set.all().order_by('text')

        self.assertEquals(result.description, "Conversation 1 UP")
        self.assertIs(result.message_set.count(), 2)
        self.assertEquals(messages[0].text, "Message 1 UP")
        self.assertEquals(messages[1].text, "Message 2 UP")

    def test_add_new_element_with_foreign_key_id(self):
        user1 = get_user_model().objects.create()
        conversation = Conversation.objects.create(description="Conversation 1", author_user=user1)
        message1 = Message.objects.create(text="Message 1", conversation=conversation, author_user=user1)
        message2 = Message.objects.create(text="Message 2", conversation=conversation, author_user=user1)

        json = {"@graph": [
            {
                "@id": "{}/messages/{}/".format(settings.BASE_URL, message1.pk),
                "text": "Message 1 UP",
                "author_user": {
                    '@id': "{}/users/{}/".format(settings.BASE_URL, user1.pk)
                }
            },
            {
                "@id": "{}/messages/{}/".format(settings.BASE_URL, message2.pk),
                "text": "Message 2 UP",
                "author_user": {
                    '@id': user1.urlid
                }
            },
            {
                "@id": "_:b1",
                "text": "Message 3 NEW",
                "author_user": {
                    '@id': user1.urlid
                }
            },
            {
                '@id': "{}/conversations/{}/".format(settings.BASE_URL, conversation.pk),
                "author_user": {
                    '@id': user1.urlid
                },
                'description': "Conversation 1 UP",
                'message_set': {
                    "@id": "{}/conversations/{}/message_set/".format(settings.BASE_URL, conversation.pk)
                }
            },
            {
                '@id': "{}/conversations/{}/message_set/".format(settings.BASE_URL, conversation.pk),
                "ldp:contains": [
                    {"@id": "{}/messages/{}/".format(settings.BASE_URL, message1.pk)},
                    {"@id": "{}/messages/{}/".format(settings.BASE_URL, message2.pk)},
                    {"@id": "_:b1"}
                ]
            }
        ]
        }

        meta_args = {'model': Conversation, 'depth': 2, 'fields': ("@id", "description", "message_set")}

        meta_class = type('Meta', (), meta_args)
        serializer_class = type(LDPSerializer)('ConversationSerializer', (LDPSerializer,), {'Meta': meta_class})
        serializer = serializer_class(data=json, instance=conversation)
        serializer.is_valid(raise_exception=True)
        result = serializer.save()

        messages = result.message_set.all().order_by('text')

        self.assertEquals(result.description, "Conversation 1 UP")
        self.assertIs(result.message_set.count(), 3)
        self.assertEquals(messages[0].text, "Message 1 UP")
        self.assertEquals(messages[1].text, "Message 2 UP")
        self.assertEquals(messages[2].text, "Message 3 NEW")

    # TODO: variation on https://git.startinblox.com/djangoldp-packages/djangoldp/issues/344
    '''def test_update_container_invalid_fk_reference_given(self):
        pass'''

    def test_save_m2m_graph_with_many_nested(self):
        invoice = {
            "@graph": [
                {
                    "@id": "./",
                    "batches": {"@id": "_:b381"},
                    "title": "Nouvelle facture",
                    "date": ""
                },
                {
                    "@id": "_:b381",
                    "tasks": {"@id": "_:b382"},
                    "title": "Batch 1"
                },
                {
                    "@id": "_:b382",
                    "title": "Tache 1"
                }
            ]
        }

        meta_args = {'model': Invoice, 'depth': 2, 'fields': ("@id", "title", "batches", "date")}

        meta_class = type('Meta', (), meta_args)
        serializer_class = type(LDPSerializer)('InvoiceSerializer', (LDPSerializer,), {'Meta': meta_class})
        serializer = serializer_class(data=invoice)
        serializer.is_valid()
        result = serializer.save()

        self.assertEquals(result.title, "Nouvelle facture")
        self.assertIs(result.batches.count(), 1)
        self.assertEquals(result.batches.all()[0].title, "Batch 1")
        self.assertIs(result.batches.all()[0].tasks.count(), 1)
        self.assertEquals(result.batches.all()[0].tasks.all()[0].title, "Tache 1")

    def test_save_m2m(self):
        skill1 = Skill.objects.create(title="skill1", obligatoire="obligatoire", slug="slug1")
        skill2 = Skill.objects.create(title="skill2", obligatoire="obligatoire", slug="slug2")

        job = {"title": "job test",
               "slug": "slug1",
               "skills": {
                   "ldp:contains": [
                       {"@id": "{}/skills/{}/".format(settings.BASE_URL, skill1.slug)},
                       {"@id": "{}/skills/{}/".format(settings.BASE_URL, skill2.slug), "title": "skill2 UP"},
                       {"title": "skill3", "obligatoire": "obligatoire", "slug": "slug3"},
                   ]}
               }

        meta_args = {'model': JobOffer, 'depth': 2, 'fields': ("@id", "title", "skills", "slug")}

        meta_class = type('Meta', (), meta_args)
        serializer_class = type(LDPSerializer)('JobOfferSerializer', (LDPSerializer,), {'Meta': meta_class})
        serializer = serializer_class(data=job)
        serializer.is_valid()
        result = serializer.save()

        self.assertEquals(result.title, "job test")
        self.assertIs(result.skills.count(), 3)
        self.assertEquals(result.skills.all()[0].title, "skill1")  # no change
        self.assertEquals(result.skills.all()[1].title, "skill2 UP")  # title updated
        self.assertEquals(result.skills.all()[2].title, "skill3")  # creation on the fly

    # variation switching the http prefix of the BASE_URL in the request
    @override_settings(BASE_URL='http://happy-dev.fr/')
    def test_save_m2m_switch_base_url_prefix(self):
        skill1 = Skill.objects.create(title="skill1", obligatoire="obligatoire", slug="slug1")

        job = {"title": "job test",
               "slug": "slug1",
               "skills": {
                   "ldp:contains": [
                       {"@id": "https://happy-dev.fr/skills/{}/".format(skill1.slug)},
                   ]}
               }

        meta_args = {'model': JobOffer, 'depth': 2, 'fields': ("@id", "title", "skills", "slug")}

        meta_class = type('Meta', (), meta_args)
        serializer_class = type(LDPSerializer)('JobOfferSerializer', (LDPSerializer,), {'Meta': meta_class})
        serializer = serializer_class(data=job)
        serializer.is_valid()
        result = serializer.save()

        self.assertEquals(result.title, "job test")
        self.assertIs(result.skills.count(), 1)
        self.assertEquals(result.skills.all()[0].title, "skill1")  # no change

    def test_save_m2m_graph_simple(self):
        job = {"@graph": [
            {"title": "job test", "slug": "slugjob",
             },
        ]}

        meta_args = {'model': JobOffer, 'depth': 2, 'fields': ("@id", "title", "skills", "slug")}

        meta_class = type('Meta', (), meta_args)
        serializer_class = type(LDPSerializer)('JobOfferSerializer', (LDPSerializer,), {'Meta': meta_class})
        serializer = serializer_class(data=job)
        serializer.is_valid()
        result = serializer.save()

        self.assertEquals(result.title, "job test")
        self.assertIs(result.skills.count(), 0)

    def test_save_m2m_graph_with_nested(self):
        skill1 = Skill.objects.create(title="skill1", obligatoire="obligatoire", slug="a")
        skill2 = Skill.objects.create(title="skill2", obligatoire="obligatoire", slug="b")

        job = {"@graph": [
            {"title": "job test",
             "slug": "slugj",
             "skills": {"@id": "_.123"}
             },
            {"@id": "_.123", "title": "skill3 NEW", "obligatoire": "obligatoire", "slug": "skill3"},
        ]}

        meta_args = {'model': JobOffer, 'depth': 2, 'fields': ("@id", "title", "skills", "slug")}

        meta_class = type('Meta', (), meta_args)
        serializer_class = type(LDPSerializer)('JobOfferSerializer', (LDPSerializer,), {'Meta': meta_class})
        serializer = serializer_class(data=job)
        serializer.is_valid()
        result = serializer.save()

        self.assertEquals(result.title, "job test")
        self.assertIs(result.skills.count(), 1)
        self.assertEquals(result.skills.all()[0].title, "skill3 NEW")  # creation on the fly

    def test_save_without_nested_fields(self):
        skill1 = Skill.objects.create(title="skill1", obligatoire="obligatoire", slug="a")
        skill2 = Skill.objects.create(title="skill2", obligatoire="obligatoire", slug="b")
        job = {"title": "job test", "slug": "c"}

        meta_args = {'model': JobOffer, 'depth': 2, 'fields': ("@id", "title", "skills", "slug")}

        meta_class = type('Meta', (), meta_args)
        serializer_class = type(LDPSerializer)('JobOfferSerializer', (LDPSerializer,), {'Meta': meta_class})
        serializer = serializer_class(data=job)
        serializer.is_valid()
        result = serializer.save()

        self.assertEquals(result.title, "job test")
        self.assertIs(result.skills.count(), 0)

    def test_save_on_sub_iri(self):
        """
            POST /job-offers/1/skills/
        """
        job = JobOffer.objects.create(title="job test")
        skill = {"title": "new SKILL"}

        meta_args = {'model': Skill, 'depth': 2, 'fields': ("@id", "title")}

        meta_class = type('Meta', (), meta_args)
        serializer_class = type(LDPSerializer)('SkillSerializer', (LDPSerializer,), {'Meta': meta_class})
        serializer = serializer_class(data=skill)
        serializer.is_valid()
        kwargs = {}
        kwargs['joboffer'] = job
        result = serializer.save(**kwargs)

        self.assertEquals(result.title, "new SKILL")
        self.assertIs(result.joboffer_set.count(), 1)
        self.assertEquals(result.joboffer_set.get(), job)
        self.assertIs(result.joboffer_set.get().skills.count(), 1)



from hashlib import md5

from Cryptodome.PublicKey import RSA
from django.db import models
from django.utils.translation import ugettext_lazy as _


class RSAKey(models.Model):

    priv_key = models.TextField(
        verbose_name=_(u'Key'), unique=True,
        help_text=_(u'Paste your private RSA Key here.'))

    class Meta:
        verbose_name = _(u'RSA Key')
        verbose_name_plural = _(u'RSA Keys')

    def __str__(self):
        return u'{0}'.format(self.kid)

    def __unicode__(self):
        return self.__str__()

    @property
    def kid(self):
        if not self.priv_key:
            return ''

        return u'{0}'.format(md5(self.priv_key.encode('utf-8')).hexdigest())

    @property
    def pub_key(self):
        if not self.priv_key:
            return ''

        _pub_key = RSA.importKey(self.priv_key).publickey()
        return _pub_key.export_key().decode('utf-8')
